@@ -6,10 +6,11 @@ require Exporter;
 use warnings;
 use strict;
 
-our $VERSION = '0.42';
+our $VERSION = '0.43';
 
-use Carp;
+use Carp 'croak';
 use Convert::Moji qw/make_regex length_one unambiguous/;
+use JSON::Parse 'json_file_to_perl';
 use utf8;
 
 our @EXPORT_OK = qw/
@@ -22,6 +23,7 @@ our @EXPORT_OK = qw/
                     circled2kana
                     circled2kanji
                     cyrillic2katakana
+		    hangul2kana
                     hira2kata
                     hw2katakana
                     is_hiragana
@@ -57,6 +59,9 @@ our @EXPORT_OK = qw/
 		    katakana2square
                     wide2ascii
 		    nigori_first
+		    hentai2kana
+		    hentai2kanji
+		    kana2hentai
 		   /;
 
 our %EXPORT_TAGS = (
@@ -1189,12 +1194,37 @@ sub kana2hangul
     my $katakana = kana2katakana ($kana);
     if (! $first2hangul) {
         load_kana2hangul ();
-        $katakana =~ s/($first2hangul_re)/$first2hangul->{$1}/g;
-        $katakana =~ s/($rest2hangul_re)/$rest2hangul->{$1}/g;
     }
+    $katakana =~ s/($first2hangul_re)/$first2hangul->{$1}/g;
+    $katakana =~ s/($rest2hangul_re)/$rest2hangul->{$1}/g;
     return $katakana;
 }
 
+my $firsth2k_re;
+my $resth2k_re;
+my $firsth2k;
+my $resth2k;
+
+sub load_hangul2kana
+{
+    load_kana2hangul ();
+    $firsth2k = { reverse %$first2hangul };
+    $resth2k = { reverse %$rest2hangul };
+    $firsth2k_re = '\b' . make_regex (keys %$firsth2k);
+    $resth2k_re = make_regex (keys %$resth2k);
+}
+
+sub hangul2kana
+{
+    my ($hangul) = @_;
+    if (! $firsth2k) {
+	load_hangul2kana ();
+    }
+    $hangul =~ s/($firsth2k_re)/$firsth2k->{$1}/;
+    $hangul =~ s/($resth2k_re)/$resth2k->{$1}/;
+    return $hangul;
+}
+
 sub kana_to_large
 {
     my ($kana) = @_;
@@ -1367,5 +1397,64 @@ sub nigori_first
     }
 }
 
+# Hentaigana (Unicode 10.0) related
+
+my $hentai_file = __FILE__;
+$hentai_file =~ s!\.pm$!/!;
+$hentai_file .= "hentaigana.json";
+# Hentai to hiragana (one to one)
+my %hen2hi;
+# Hiragana to hentai (one to many)
+my %hi2hen;
+# Hentaigana to kanji
+my %hen2k;
+# Hentai regex
+my $hen_re;
+# Hiragana to hentai regex. This is slightly complex because a single
+# hentaigana might correspond to two or more hiraganas.
+my $hi2hen_re;
+# Hentai data
+my $hendat;
+
+sub load_hentai
+{
+    $hendat = json_file_to_perl ($hentai_file);
+    for my $h (@$hendat) {
+	my $hi = $h->{hi};
+	my $hen = chr ($h->{u});
+	$hen2hi{$hen} = $hi;
+	push @{$hi2hen{$hi}}, $hen;
+	$hen2k{$hen} = $h->{ka};
+    }
+    $hen_re = make_regex (keys %hen2hi);
+    $hi2hen_re = make_regex (keys %hi2hen);
+}
+
+sub hentai2kana
+{
+    my ($text) = @_;
+    if (! $hendat) {
+	load_hentai ();
+    }
+    $text =~ s/$hen_re/$hen2hi{$1}/g;
+    return $text;
+}
+
+sub kana2hentai
+{
+    my ($kana) = @_;
+    # what to do?
+}
+
+sub hentai2kanji
+{
+    my ($text) = @_;
+    if (! $hendat) {
+	load_hentai ();
+    }
+    $text =~ s/$hen_re/$hen2k{$1}/g;
+    return $text;
+}
+
 1; 
 
